#!/usr/bin/env perl

# Enable Perl warnings
use strict;
use warnings;
use Term::ANSIColor;	# Module for colors and errors
use Getopt::Long qw(:config no_ignore_case bundling);

#1 Define transformation fields
our (%g)=	('TCA'=>'S','TCC'=>'S','TCG'=>'S','TCT'=>'S','TTC'=>'F',
			'TTT'=>'F','TTA'=>'L','TTG'=>'L','TAC'=>'Y','TAT'=>'Y',
			'TAA'=>'_','TAG'=>'_','TGC'=>'C','TGT'=>'C','TGA'=>'_',	# Stop codons => TAA, TAG, TGA => Z
			'TGG'=>'W','CTA'=>'L','CTC'=>'L','CTG'=>'L','CTT'=>'L',
			'CCA'=>'P','CCC'=>'P','CCG'=>'P','CCT'=>'P','CAC'=>'H',
			'CAT'=>'H','CAA'=>'Q','CAG'=>'Q','CGA'=>'R','CGC'=>'R',
			'CGG'=>'R','CGT'=>'R','ATA'=>'I','ATC'=>'I','ATT'=>'I',
			'ATG'=>'M','ACA'=>'T','ACC'=>'T','ACG'=>'T','ACT'=>'T',
			'AAC'=>'N','AAT'=>'N','AAA'=>'K','AAG'=>'K','AGC'=>'S',
			'AGT'=>'S','AGA'=>'R','AGG'=>'R','GTA'=>'V','GTC'=>'V',
			'GTG'=>'V','GTT'=>'V','GCA'=>'A','GCC'=>'A','GCG'=>'A',
			'GCT'=>'A','GAC'=>'D','GAT'=>'D','GAA'=>'E','GAG'=>'E',
			'GGA'=>'G','GGC'=>'G','GGG'=>'G','GGT'=>'G',
			'UCA'=>'S','UCC'=>'S','UCG'=>'S','UCU'=>'S','UUC'=>'F',
			'UUU'=>'F','UUA'=>'L','UUG'=>'L','UAC'=>'Y','UAU'=>'Y',
			'UAA'=>'_','UAG'=>'_','UGC'=>'C','UGU'=>'C','UGA'=>'_',
			'UGG'=>'W','CUA'=>'L','CUC'=>'L','CUG'=>'L','CUU'=>'L',
			'CCU'=>'P','CAU'=>'H','CGU'=>'R','AUA'=>'I','AUC'=>'I',
			'AUU'=>'I','AUG'=>'M','ACU'=>'T','AAU'=>'N','AGU'=>'S',
			'GUA'=>'V','GUC'=>'V','GUG'=>'V','GUU'=>'V','GCU'=>'A',
			'GAU'=>'D','GGU'=>'G'
);

our (%np) = (	'A'=>'T','T'=>'A','G'=>'C','C'=>'G',
				'U'=>'A','N'=>'N','R'=>'R','Y'=>'Y',
                'M'=>'M','K'=>'K','S'=>'S','W'=>'W',
				'B'=>'B','D'=>'D','H'=>'H','V'=>'V'
);

#2 Define PRIMARY FIELD and VARIABLES
our @gene = "";				# GLOBAL FIELD - for store CDS records
our $gene_counter = 0;		# Counter for CDS RECORDS
our @mutation_data;			# GLOBAL FIELD - for store MUTATION DATA
our $mutation_index = 0;	# Counter for MUTATION RECORDS
our $pocitadlo = 0;			# Counter for processing 
our @history_transcript;	# GLOBAL FIELD - for already processed transcripts
our $history_index = 0;		# Index for processed transcripts

our $STOP = 0;			# Count of STOP codons in sequence
our $SHIFTER = 0;		# Count for shift SEQ if multiple IND/DEL
our $flag_INS = 0;		# Flag for insertion
our $flag_DEL = 0;		# Flag for deletion

# Names of output files
our $file_name_WT_nucl = "WT_nucl.nucl";
our $file_name_WT_pep = "WT_pep.pep";
our $file_name_MUT_nucl = "MUT_nucl.nucl";
our $file_name_MUT_pep = "MUT_pep.pep";
our $file_name_log = "logfile.log";
our $input_reference = "";
my $input_GTF = "";
my $input_MUT = "";
my $help = 0;

# INTERFACE
# Print usage and exit if requested - NAPOVEDA; Kontrola poctu parametru.
if (scalar(@ARGV) == 0) 
{
	print STDERR color("bold red"), "ERROR: Bad number of arguments.\n\n", color("reset");
	&print_usage();
	exit(0);
}

###################################################################################################
########################################### OPTIONS ###############################################
###################################################################################################

# Extract parameters values
GetOptions(
	'out_WT_nucl=s'		=> \$file_name_WT_nucl,			# Output - WT nucleotide sequence
	'out_MUT_nucl=s'	=> \$file_name_MUT_nucl,		# Output - MUT nucleotide sequence
	'out_WT_pep=s'		=> \$file_name_WT_pep,			# Output - WT peptide sequence
	'out_MUT_pep=s'		=> \$file_name_MUT_pep,			# Output - MUT peptide sequence
	'out_log=s'			=> \$file_name_log,				# Output - Logfile name
	'input_reference=s'	=> \$input_reference,			# Input - file for reference in fasta format
	'input_GTF=s'		=> \$input_GTF,					# Input - file with GTF
	'input_MUT=s'		=> \$input_MUT,					# Input - file with mutation info
	'help'				=> \$help						# For print help message
);

# Help message
if ($help)
{
	&print_usage();
	exit(0);
}

# Verify parameters (Input params must be included.)
if (defined($input_reference) && defined($input_GTF) && defined($input_MUT) && (!($input_reference eq "")) && (!($input_reference eq "")) && (!($input_reference eq "")))
{
	print STDERR "Program is launched with following parameters:\n\nout_WT_nucl\t$file_name_WT_nucl\nout_MUT_nucl\t$file_name_MUT_nucl\nout_WT_pep\t$file_name_WT_pep\n";
	print STDERR "out_MUT_pep\t$file_name_MUT_pep\nout_log\t$file_name_log\ninput_reference\t$input_reference\ninput_GTF\t$input_GTF\ninput_MUT\t$input_MUT\n\n";
}
else
{
	print STDERR color("bold red"), "ERROR: Format of input parameters is not supported.\n", color("reset");
}

#3 READ GTF FILE
if (&read_GTF_file($input_GTF) == 0) {exit (0);}


#4 Reset files
&files_reset();

#4 READ MUTATION FILE
if (&read_mutation_file($input_MUT) == 0) {exit (0);}

#6 Print head to log_file
&mutation_info_print("Gene name","Transcript name","Position","WT peptide","Mutation peptide");

#7 PROCESS OF MUTATIONS
while ($pocitadlo < $mutation_index)
{
	&search_trans($pocitadlo);
	$pocitadlo++;
}

exit (0);

#######################################################################################
########################### SUB FUNCTIONS #############################################
#######################################################################################

#READ 1# - Read GTF file
#Return: 0 if zero lines, 1 if nonzero lines
sub read_GTF_file
{
	#1 Define variables
	my $file_name = shift;					#1 Name of input file
	#@gene - GLOBAL FIELD FOR GENES
	my $line = "";							#1 Line for read file
	#$gene_counter - GLOBAL VARIABLE FOR CONTING OF GENES
	my @sloupce;							#1 Field for 
	my @sub_sloupce;

	#2 Open GTF file
	if (!(open (ZDROJ, $file_name))) 
	{
		#A Print warning
		print STDERR color("red"), "ERROR: Input file cannot be open.\n", color("reset");
		#B Exit program
		exit (0);
	}

	#3 Go file until you get RECORDS
	while ($line = <ZDROJ>) 
	{
		if (!($line =~ m/#/)) 
		{
			last;
		}
	}

	#4 READ GTF data
	while ($line = <ZDROJ>)
	{
		@sloupce = split("\t", $line);
		if ($sloupce[2] eq "CDS")
		{
			#A Increment gene counter
			$gene_counter++;
			#B Get data into GLOBAL FIELD
			$gene[$gene_counter][1] = $sloupce[0];	# Chromosome
			$gene[$gene_counter][2] = $sloupce[2];	# Source of SEQ (CDS)
			$gene[$gene_counter][3] = $sloupce[3];	# Start
			$gene[$gene_counter][4] = $sloupce[4];	# Stop
			$gene[$gene_counter][5] = $sloupce[6];	# Frame
			$gene[$gene_counter][6] = $sloupce[8];	# Description

			#C Split description into sub-collumns
			@sub_sloupce = split(";", $sloupce[8]);

			#D Reset collumns from field sloupce
			$sloupce[0] = "";
			$sloupce[1] = "";
			$sloupce[2] = "";
			$sloupce[3] = "";
			$sloupce[4] = "";
			$sloupce[5] = "";
			$sloupce[6] = "";
			$sloupce[7] = "";
			$sloupce[8] = "";

			#E Get data from subcollumns
			$gene[$gene_counter][7] = $sub_sloupce[0];	# Gene identification
			$gene[$gene_counter][8] = $sub_sloupce[1];	# Gene version
			$gene[$gene_counter][9] = $sub_sloupce[2];	# Transcript identification
			$gene[$gene_counter][10] = $sub_sloupce[3];	# Transcript version
			$gene[$gene_counter][11] = $sub_sloupce[4];	# Exon number
			$gene[$gene_counter][12] = $sub_sloupce[5];	# Gene name

			#F Get gene identificator
			@sub_sloupce = split('"', $gene[$gene_counter][7]);
			$gene[$gene_counter][7] = $sub_sloupce[1];	# Gene identificator
			$sub_sloupce[1] = "";

			#G Get gene version
			@sub_sloupce = split('"', $gene[$gene_counter][8]);
			$gene[$gene_counter][8] = $sub_sloupce[1];	# Gene version
			$sub_sloupce[1] = "";

			#H Get transcript identification
			@sub_sloupce = split('"', $gene[$gene_counter][9]);
			$gene[$gene_counter][9] = $sub_sloupce[1];	# Transcript identification
			$sub_sloupce[1] = "";

			#I Get transcript version
			@sub_sloupce = split('"', $gene[$gene_counter][10]);
			$gene[$gene_counter][10] = $sub_sloupce[1];	# Transcript version
			$sub_sloupce[1] = "";

			#J Get EXON number
			@sub_sloupce = split('"', $gene[$gene_counter][11]);
			$gene[$gene_counter][11] = $sub_sloupce[1];	# Exon number
			$sub_sloupce[1] = "";
	
			#K Get GENE NAME
			@sub_sloupce = split('"', $gene[$gene_counter][12]);
			$gene[$gene_counter][12] = $sub_sloupce[1];	# Gene name
			$sub_sloupce[1] = "";

			#L PRINT COUNTER MESSAGE
			if ($gene_counter%100000 == 0)
			{
				print STDERR "$gene_counter lines were processed from GTF file.\n"; 
			}
		}
	}
	#5 Print out message
	print STDERR "$gene_counter lines were processed from GTF file.\n";
	#6 Return
	if ($gene_counter > 10)
	{
		return 1;
	}
	else
	{
		print STDERR color("bold red"), "ERROR: Input GTF file contains less than 10 lines. Please check the input file.\n", color("reset");
		return 0;
	}
}

#READ 2# - Read mutation file
#Return: 0 if zero lines, >0 if nonzero lines
sub read_mutation_file
{
	#1 Define local variables
	my $file_name = shift;		# Input file with mutation
	my $line;					# Line for parsing records			
	my @sloupce;				# Field for 
	#mutation_index - for counting number of mutations
	#mutation_data - GLOBAL field for store data for mutation

	#2 Open MUTATION FILE
	if (!(open (MUTA, $file_name))) 
	{
		#A Print warning
		print STDERR color("red"), "ERROR: Input file cannot be open.\n", color("reset");
		#B Exit program
		exit (0);
	}

	#3 READ MUTATION DATA AND TESTING
	while ($line = <MUTA>)
	{
		#A Reset field
		$sloupce[0] = "";
		$sloupce[1] = "";
		$sloupce[2] = "";
		$sloupce[3] = "";

		#B Split line
		@sloupce = split("\t", $line);

		#C Preformated input
		$sloupce[0] =~ s/\n//g;
		$sloupce[1] =~ s/\n//g;
		$sloupce[2] =~ s/\n//g;
		$sloupce[3] =~ s/\n//g;
		$sloupce[0] =~ s/ //g;
		$sloupce[1] =~ s/ //g;
		$sloupce[2] =~ s/ //g;
		$sloupce[3] =~ s/ //g;
		
		#D1 Verify collumns values
		if ($sloupce[1] > 0 && (!($sloupce[0] eq "")) && (!($sloupce[2] eq "")) && (!($sloupce[3] eq "")) &&
			(!($sloupce[2] =~ m/\//)) && (!($sloupce[3] =~ m/\//)) && (!($sloupce[2] =~ m/\*/)) && (!($sloupce[3] =~ m/\*/)))
		{
			# Insert data into GLOBAL FIELD
			$mutation_data[$mutation_index][0] = $sloupce[0];
			$mutation_data[$mutation_index][1] = $sloupce[1];
			$mutation_data[$mutation_index][2] = $sloupce[2];
			$mutation_data[$mutation_index][3] = $sloupce[3];

			# Increment index
			$mutation_index++;
		}
		else
		#D2 Error line detected
		{
			# Open errorlog FILE
			open(ERRLOG, '>>', "error.log");

			# Print data into file
			print ERRLOG $sloupce[0], "\t", $sloupce[1], "\t", $sloupce[2], "\t", $sloupce[3], "\n";

			# Close file and return
			close ERRLOG;
		}
	}

	#4 VERIFY READING
	print STDERR color("yellow"), "INFO: For translation were loaded $mutation_index mutations.\n", color("reset");
	print STDERR color("yellow"), "Error lines are reported in error.log file.\n\n", color("reset");

	#5 SORT ARRAY
	@mutation_data = sort { $a->[0] cmp $b->[0] || $a->[1] <=> $b->[1] } @mutation_data;
	# Print array
	#use Data::Dump; dd \@mutation_data;

	#6 Return value
	return $mutation_index;
}

#SEARCH 3# - Search if mutation is in exon and print message
# SEARCH BASED ON MUTATION POSITION
sub search_trans
{
	#1 Extract variables
	my $mutation_index = shift;			# Index for global field mutation_data contains mutations
	#Global FIELD - $gene[$i][$x] with CDS
	#Global VARIABLE - $gene_counter
	#Global FIELD - $mutation_data[$i][$x] with mutations from input file
	#history_transcript - GLOBAL field for store history mutation
	my @search_field;					# Transcript name for search
	my $search_index = 1;				# Index fot store transcripts
	@search_field[$search_index] = "";	# Field initialization
	my $chrom_switch = 0;				# Chromosome switcher
	my $control = 1;					# Control for while loop
	my $i = 0;							# Counter for search
	my $local_history = "";				# Local history variable
	my $local_history_index = 1;		# Local history index
	my $no_search = 1;					# No search switch
	my $x = 1;

	#2 FIND TRANSCRIPTS IF EXIST
	while ($control)
	{
		$i++;
		# If search FAILED
		if ($i > $gene_counter) { return 0;	}		
		# SEARCH PROCES
		if (($gene[$i][1] eq $mutation_data[$mutation_index][0]) && ($gene[$i][3] <= $mutation_data[$mutation_index][1]) && ($gene[$i][4] >= $mutation_data[$mutation_index][1]))
		{
			# Write new CDS into FIELD
			if ((!($search_field[$search_index] eq $gene[$i][9])) && (!($search_field[$search_index] eq "")))
			{
				$search_index++;
				$search_field[$search_index] = $gene[$i][9];
				$search_index++;
				$search_field[$search_index] = "";
			}
			elsif ((!($search_field[$search_index] eq $gene[$i][9])) && ($search_field[$search_index] eq ""))
			{
				$search_field[$search_index] = $gene[$i][9];
			}

		}
		# SWITCH CHROM SWITCH
		if ($gene[$i][1] eq $mutation_data[$mutation_index][0]) { $chrom_switch = 1;}
		# SWITCH CONTROL
		if ((!($gene[$i][1] eq $mutation_data[$mutation_index][0])) && $chrom_switch > 0) {last;}
	}

	#3 EXTRACT DATA
	if ($search_field[$search_index] eq "") {$search_index--;}
	if ($search_index == 0) {return 0;}
	$i = 0;
	$x = 0;
	
	#4 WHILE loop for extract
	while ($i < $search_index)
	{
		# Increment search index
		$i++;
		$x = 0;
		# Verify TRANSCRIPT ID - If not present insert
		while ($x < $gene_counter) 
		{
			if (defined($history_transcript[$x]) && ($history_transcript[$x] eq $search_field[$i]))
			{
				$no_search = 0;
				last;	
			}
			elsif (defined($history_transcript[$x]))
			{
				$x++;
			}
			elsif (!(defined($history_transcript[$x])))
			{
				last;
			}
		}

		if ($no_search > 0)
		{
			# DO SEARCHING
			&data_extraction($search_field[$i],$mutation_index);
			# STORE NEW TRANSCRIPT
			$history_transcript[$x] = $search_field[$i];
		}
		else
		{
			$no_search = 1;
		}
	}

	#5 Return value after search
	return 0;
}	
	
# EXTRACT DATA ACCORDING TO MUTATION
sub data_extraction
{
	#1 Extract variables a define
	my $trans_id = shift;		# Query transcript ID
	my $m_index = shift;		# Query mutation ID
	my $i = 0; 					# Counter for WHILE
	my $control = 1;			# Control variable fo WHILE
	# gene_counter - global counter
	# gene - GLOBAL field
	my @sub_field;				# LOCAL FIELD for CDS records for transcript
	## TR CHARACTERS
	my $chrom = 0;
	my $frame = 0;
	my $gene_id = "";
	my $gene_name = "";
	
	my $line;					# Line for parsing
	my $text = "";				# Text for WT sequence
	my $local_MUT = "";			# Temp mutation sequence
	my $dx = 1;					# Counter for multiple mutations
	my $data = "";				# Data for WT SEQ
	my $mut = "";				# Data for MUT SEQ
	my $head = "";				# Head of the sequence

	#2 FIND TRANSCRIPT in global field ($gene)
	while ($control && $i < $gene_counter)
	{
		#A Increment counter
		$i++;
		#B If Transcript is found
		if ($gene[$i][9] eq $trans_id)
		{	
			# Extract value for LOCAL FIELD
			$sub_field[$control][0] = $gene[$i][3];	# start exon
			$sub_field[$control][1] = $gene[$i][4];	# stop exon
			$sub_field[$control][2] = $gene[$i][11]; # Exon number

			# Extract values for transcript
			$chrom = $gene[$i][1];		# CHROMOSOME
			$frame = $gene[$i][5];		# FRAME
			$gene_id = $gene[$i][7];	# GENE_ID
			$gene_name = $gene[$i][12];	# GENE_NAME

			# Increment conter for TRANSCRIPT
			$control++;
		}
		else
		{
			# End cycle
			if ($control > 2) {last;}
		}
	}

	#3 FOR LOOP FOR GETTING SEQUENCE FROM EXONS
	for($i=1; $i<$control; $i++)
	{
		#A Get SAMTOOLS order and parse
		open(SAMTOOLS,"samtools faidx $input_reference $chrom:$sub_field[$i][0]-$sub_field[$i][1] |");
    	while($line = <SAMTOOLS>)
		{
			if (!($line =~ m/>/))
			{
				$line =~ s/\n//g;	# Remove /n from name of sequence
				$text = $text . $line;
			}
    	}
    	close(SAMTOOLS);
		$text =~ s/\n//g;

		#B1 MUTATION DETECTED IN EXON
		if (($mutation_data[$m_index][1] >= ($sub_field[$i][0])) && ($mutation_data[$m_index][1] <= ($sub_field[$i][1])))
		{
			#A Set shifter
			$SHIFTER = 0;
			#B First MUTATION modification
			$local_MUT = (&modif_seq($text,$mutation_data[$m_index][1]-$sub_field[$i][0],$mutation_data[$m_index][2],$mutation_data[$m_index][3]));

			#C While loop for multiple mutation
			while ($dx > 0)
			{
				#c1 IF chromosome is the same and mutation is in RANGE of exon
				if (exists($mutation_data[$m_index][0]) && exists($mutation_data[$m_index+1][0]) &&
					($mutation_data[$m_index][0] eq $mutation_data[$m_index+1][0]) && 
					($mutation_data[$m_index+1][1] >= ($sub_field[$i][0])) && 
					($mutation_data[$m_index+1][1] <= ($sub_field[$i][1])))
				{
					$m_index++;	# Increase mutation index
					$dx++;		# Increase dx index
					# Do mutation for already mutation sequence
					$local_MUT = (&modif_seq($local_MUT,$mutation_data[$m_index][1]-$sub_field[$i][0],$mutation_data[$m_index][2],$mutation_data[$m_index][3]));
				}
				#c2 IF chromosome is the same but mutation is in next EXON
				elsif (exists($sub_field[$i+1][0]) && exists($mutation_data[$m_index+1][0]) &&
						($mutation_data[$m_index][0] eq $mutation_data[$m_index+1][0]) &&
						($mutation_data[$m_index+1][1] >= ($sub_field[$i+1][0])) && 
						($mutation_data[$m_index+1][1] <= ($sub_field[$i+1][1])))
				{
					$m_index++;	# Increase mutation index
					$dx++;		# Increase dx index for decrement
					last;
				}
				#c3 If other mutation is not present
				else
				{
					last;
				}
			}

			#D Add exon sequence/mutation sequence to whole sequence
			if ($frame eq "+")
			{
				$data = $data . $text;
				$mut = $mut . $local_MUT;
			}
			elsif ($frame eq "-")
			{
				$data = $data . (&reverse_seq($text));
				$mut = $mut . (&reverse_seq($local_MUT));
			}
		}
		#B2 MUTATION IS NOT DETECTED IN EXON
		else
		{
			# Add sequence to whole sequence
			if ($frame eq "+")
			{
				$data = $data . $text;
				$mut = $mut . $text;
			}
			elsif ($frame eq "-")
			{
				$data = $data . (&reverse_seq($text));
				$mut = $mut . (&reverse_seq($text));
			}
		}
		$text = "";
	}

	#4 RESET COUNTER FOR MULTIPLE MUTATION
	$m_index = $m_index + 1 - $dx;
	$dx = 0;

	#5 PRINT AND VERIFY RESULT
	#5A Verify if mutation has an peptide impact and return logic value
	$STOP = 0;
	&translate_seq($data);
	if ($STOP == 0)
	{
		$head = '';
		# If in WT seq is not STOP CODON detect mutation
		$head = &mut_in_prot($gene_name,$trans_id,&translate_seq($data),&translate_seq($mut));
		# Format head text
		$head = $gene_name . "_" . $trans_id . $head;
		# Print WT NC and PEP on files
		&sequence_print($file_name_WT_nucl,$head . "_WT",$data,60);
		&sequence_print($file_name_WT_pep,$head . "_WT",&translate_seq($data),60);
		# Print MUT NC and PEP on files
		&sequence_print($file_name_MUT_nucl,$head . "_MUT",$mut,60);
		&sequence_print($file_name_MUT_pep,$head . "_MUT",&translate_seq($mut),60);
	}
	else
	{
		$head = '';
		# If in WT seq is STOP codon detected give only nucl sequence and print ERROR
		&mutation_info_print($gene_name,$trans_id,0,"Stop codon in WT sequence","Stop codon in WT sequence");
		print STDERR color("bold red"), "ERORR: Stop codon detected in WT protein sequence for transcript $trans_id. Mistake is probably in gene annotation file.\n", color("reset");
		# Format head text
		$head = $gene_name . "_" . $trans_id . "_Stop_IN_WT";
		# Print WT NC and PEP on files
		&sequence_print($file_name_WT_nucl,$head . "_WT",$data,60);
		&sequence_print($file_name_MUT_nucl,$head . "_MUT",$mut,60);
	}

	#5 Exit
	$flag_INS = 0;
	$flag_DEL = 0;
	return 0;
}


# PRINT SEQ IN DEFINED FORMAT
sub print_seq
{
	# Extract variables
	my $data = shift;
	my $freq = shift;
	my $res = "";
	my $part = "";
	my $i = 0;
	# Modify input string
	$data =~ s/\n//g;	# Remove /n from name of sequence
	# Cut sequence
	for($i=0; $i<(length($data)); $i+=$freq)
	{
		#A Get substring for codon
		$part = substr($data,$i,$freq);
		#B Print the result
		print $part, "\n";
		$res = $res . $part . "\n";
	}
	# Return value or return text string
	return $res;
}


# TRANSLATE SEQUENCE TO PEPTIDE
sub translate_seq 
{
	#1 Define local variables
	my $sequence = shift;		# Sequence
	my $i = 0;					# For loop counter
	my $protein = '';			# TEMP sequence
	my $codon = '';
	my $stop_flag = 0;
	
	#2 Translate tricodons
	$sequence =~ s/\n//g;
	$STOP = 0;
	for($i=0; $i<(length($sequence)-2); $i+=3)
	{
		#A Get substring for codon
		$codon = substr($sequence,$i,3);
		$codon = uc $codon;

		#B Get translate codon
		if(exists $g{$codon})
		{
			if ($g{$codon} eq "_") 
			{
				$STOP++;
				$stop_flag = 1;
				$protein = $protein . ($g{$codon});
			}
			if ($stop_flag == 0) {$protein = $protein . ($g{$codon});}
		}
		else
		{
			print STDERR color("bold red"), "Bad codon \"$codon\" detected !!\n", color("reset");
		}	
	}	
	return $protein;
}

# REVERSE SEQUENCE FRAME
sub reverse_seq 
{
	#1 Get local variable and define
	my $in_seq = shift;
	my $a = 0;				# Pocitadlo
	my $nucl = "";
	my $rever = "";

	#2 For loop for translation
	$in_seq =~ s/\n//g;
	for ($a=0; $a<length($in_seq);$a+=1)
	{
		#A Get nucleotide
		$nucl = substr($in_seq,$a,1);
		$nucl = uc $nucl;
		#B Give reverse nucleotide
		if (exists $np{$nucl})
		{
			$rever = ($np{$nucl}) . $rever;
		}
		else
		{
			print STDERR color("bold red"), "Unknown nucleotide detected \"$nucl\"!!\n", color("reset");
		}
	}
	
	#3 Return result
	return $rever;
}

#MODIFY SEQ TO MUT
sub modif_seq
{
	#1 Get local variables
	my $in_seq = shift;
	my $position = shift;
	my $modify = shift;
	my $new = shift;
	my $result = "";
	my $insert = "";

	#2 Modify NEW SEQ
	$position = $position + $SHIFTER;
	$new =~ s/\n//g;		
	$new =~ s/ //g;
	$modify =~ s/\n//g;		
	$modify =~ s/ //g;

	#3 Get sequence before modification
	$result = substr($in_seq,0,$position);
	$insert = $new;
	

	#4A Detection SNP - substitution
	if (length($new) == 1)
	{
		# SNP detected
		if ((substr($in_seq,$position,1)) eq $modify)
		{
			$result = $result . $new;
			$result = $result . (substr($in_seq,$position+1,(length($in_seq)-$position)));
			return $result;
		}
		elsif (!((substr($in_seq,$position,1)) eq $modify))
		{
			print color("bold yellow"), "WARRING: Reference nucleotide is not equal with the modify nucleotide for SNP event. Reference nucleotide is ", substr($in_seq,$position,1), " and nucleotide in mutation file for modification is $modify\n", color("reset");
			$result = $result . $new;
			$result = $result . (substr($in_seq,$position+1,(length($in_seq)-$position)));
			return $result;
		}
		else
		{
			print color("bold red"), "ERROR: Unknown error in modify sequence for SNP event.\n", color("reset");
			return undef;
		}
	}
	#4B Detection INSERTION
	elsif ((substr($new,0,1)) eq "+")
	{
		# INS detected
		$new = (substr($new,1,length($new)-1));
		$SHIFTER = $SHIFTER + (length($new));
		$flag_INS = $flag_INS + length($new);
		if ((substr($in_seq,$position,1)) eq $modify)
		{
			$result = $result . (substr($in_seq,$position,1)) . $new;
			$result = $result . (substr($in_seq,$position+1,(length($in_seq)-$position)));
			return $result;
		}
		elsif (!((substr($in_seq,$position,1)) eq $modify))
		{
			print color("bold yellow"), "WARRING: Reference nucleotide is not equal with the modify nucleotide for INSERTION event. Reference nucleotide is ", substr($in_seq,$position,1), " and nucleotide in mutation file is $modify\n", color("reset");
			$result = $result . (substr($in_seq,$position,1)) . $new;
			$result = $result . (substr($in_seq,$position+1,(length($in_seq)-$position)));
			return $result;
		}
		else
		{
			print color("bold red"), "ERROR: Unknown error in modify sequence for INSERTION event.\n", color("reset");
			return undef;
		}
	}
	#4C Detection DELETION
	elsif ($new =~ m/-/)
	{
		# DEL detected
		$new =~ s/-//g;
		$SHIFTER = $SHIFTER - (length($new));
		$flag_DEL = $flag_DEL + length($new);
		if (((substr($in_seq,$position,1)) eq $modify) && (substr($in_seq,$position+1,length($new)) eq $new))
		{
			$result = $result . (substr($in_seq,$position,1));
			$result = $result . (substr($in_seq,$position+1+(length($new)),(length($in_seq)-$position-(length($new)))));
			return $result;
		}
		elsif ((!((substr($in_seq,$position,1)) eq $modify)) ||	(!(substr($in_seq,$position+1,length($new)) eq $new)))
		{
			print color("bold yellow"), "WARRING: Reference nucleotide is not equal with the modify nucleotide for DELETION event. Reference is ", substr($in_seq,$position,1), " and modify is $modify\n", color("reset");
			$result = $result . (substr($in_seq,$position,1));
			$result = $result . (substr($in_seq,$position+1+(length($new)),(length($in_seq)-$position-(length($new)))));
			return $result;
		}
		else
		{
			print color("bold red"),"ERROR: Unknown error in modify sequence for DELETION event.\n",color("reset");
			return undef;
		}
	}
	#4D UNKNOWN ERROR
	else
	{
		print color("bold red"),"ERROR: Unknown error in modify sequence. New seq is: $new\n",color("reset");
		return undef;
	}
}

#SEARCH MUTATION IN SEQ
sub mut_in_prot
{
	#1 Get input data
	my $g_name = shift;
	my $t_name = shift;
	my $WT_string = shift;
	my $MUT_string = shift;
	my $a = 0; # Pocitadlo
	my $i = 0; # INS/DEL pocitadlo
	my $i_max = 0;
	my $score = 0; #INS/DEL score
	my $x = 0;
	my $m_count = "";
	# SEARCH
	my $b = 0;
	my $inter_shift = 0;

	#2 Preprocessing - remove no characters
	$WT_string =~ s/\n//g;
	$MUT_string =~ s/\n//g;
	$WT_string =~ s/ //g;
	$MUT_string =~ s/ //g;

	for ($a=0; $a<length($WT_string);$a+=1)
	{
		if ($a == length($MUT_string)) {return $m_count;}
		# Mutation is in peptide so print message into logfile
		if (!(substr($WT_string,$a+$inter_shift,1) eq (substr($MUT_string,$a,1))))
		{	
			#A If the peptide has STOP CODON print message and cut the cycle
			if (substr($MUT_string,$a,1) eq "_")
			{
				print STDERR color("green"), "Mutation with STOP CODON is on possition ", $a+1, " at gene name $g_name and transcript $t_name where is ", substr($WT_string,$a+$inter_shift,1), " mutated to STOP codon.\n", color("reset");
				&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),"Stop");
				$x = $a + 1;
				$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . "Stop";
			}
			#B If INSERTION
			elsif ($flag_INS > 0)
			{
				#B1 Set maxcounter i value
				if ($a+11 < length($MUT_string)) {$i_max = 20;}
				else {$i_max = length($MUT_string) - $a - 1;}
				#print $a, "\t", length($MUT_string), "\n";
				$i = 0;
				$b = 0;
				$score = 0;
 
				#B2 Detection frameshift
				for ($i=-1; $i<$i_max; $i+=1)
				{
					if (substr($WT_string,$a+$b+$inter_shift,1) eq (substr($MUT_string,$a+$i,1)) || substr($WT_string,$a+$b+$inter_shift+1,1) eq (substr($MUT_string,$a+$i+1,1))) 
					{
						$score++; 
						$b++;
					}
				}

				#B3 Score is less than 5 - frame shift detected
				if ($score < 5)
				{
					# Frameshift is detected
					print STDERR color("green"), "Insertion event on possition ", $a+1, " at gene name $g_name and transcript $t_name detected with FRAMESHIFT.\n", color("reset");
					&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),"Insertion frameshift");
					$x = $a + 1;
					$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . "InsertionFrameshift";
					return $m_count;
				}
				else
				{
					# Detection MUTATION/INSERTION
					$i_max = $flag_INS/3 + 1;
					$score = 0; 
					for ($i=0; $i<$i_max; $i+=1)
					{
						if (substr($WT_string,$a+$inter_shift,1) eq (substr($MUT_string,$a+$i,1))) {$score = 1;}
					}
					# Print message
					if ($score > 0)
					{
						# Frameshift is not detected - standard operations
						print STDERR "Insertion event on possition ", $a+1, " at gene name $g_name and transcript $t_name detected.\n";
						&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),"+" . (substr($MUT_string,$a,1)));
						$x = $a + 1;
						$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . (substr($MUT_string,$a,1));
						$inter_shift--;
						$score = 0;
					}
					else
					{
						# Frameshift is not detected - standard operations
						print STDERR "Mutation is on possition ", $a+1, " at gene name $g_name and transcript $t_name where is ", substr($WT_string,$a+$inter_shift,1), " mutated to ", substr($MUT_string,$a,1), ".\n";
						&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),(substr($MUT_string,$a,1)));
						$x = $a + 1;
						$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . (substr($MUT_string,$a,1));
						$score = 0;
					}
				}
			}
			#C If DELETION
			elsif ($flag_DEL > 0)
			{
				#C1 Set maxcounter i value
				if ($a+11 < length($MUT_string)) {$i_max = 20;}
				else {$i_max = length($MUT_string) - $a - 1;}
				$i = 0;
				$b = 0;
				$score = 0;

				#B2 Detection frameshift
				for ($i=-1; $i<$i_max; $i+=1)
				{
					if (substr($WT_string,$a+$i+$inter_shift,1) eq (substr($MUT_string,$a+$b,1)) || substr($WT_string,$a+$i+$inter_shift+1,1) eq (substr($MUT_string,$a+$b+1,1))) 
					{
						$score++; 
						$b++;
					}
				}

				#B3 Score is less than 5 - frame shift detected
				if ($score < 5)
				{
					# Read frame is DETECTED
					print STDERR color("green"), "Deletion event on possition ", $a+1, " at gene name $g_name and transcript $t_name detected with FRAMESHIFT.\n", color("reset");
					&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),"DeletionFrameshift");
					$x = $a + 1;
					$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . "DeletionFrameshift";
					return $m_count;
				}
				else
				{
					# Detection MUTATION/DELETION
					$i_max = $flag_DEL/3 + 1;
					$score = 0; 
					for ($i=0; $i<$i_max; $i+=1)
					{
						if (substr($WT_string,$a+$inter_shift+$i,1) eq (substr($MUT_string,$a,1))) {$score = 1;}
					}
					# Print message
					if ($score > 0)
					{
						# Frameshift is not detected - standard opertations
						print STDERR "Deletion event on possition ", $a+1, " at gene name $g_name and transcript $t_name detected.\n";
						&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),"Deletion");
						$x = $a + 1;
						$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . "Deletion";
						$inter_shift++;
						$score = 0;
					}
					else
					{
						print STDERR "Mutation is on possition ", $a+1, " at gene name $g_name and transcript $t_name where is ", substr($WT_string,$a+$inter_shift,1), " mutated to ", substr($MUT_string,$a,1), ".\n";
						&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),(substr($MUT_string,$a,1)));
						$x = $a + 1;
						$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . (substr($MUT_string,$a,1));
						$score = 0;
					}
				}
			}
			#D If cause SNP
			else
			{
				print STDERR "Mutation is on possition ", $a+1, " at gene name $g_name and transcript $t_name where is ", substr($WT_string,$a+$inter_shift,1), " mutated to ", substr($MUT_string,$a,1), ".\n";
				&mutation_info_print($g_name,$t_name,$a+1,(substr($WT_string,$a+$inter_shift,1)),(substr($MUT_string,$a,1)));
				$x = $a + 1;
				$m_count = $m_count . "_" . (substr($WT_string,$a+$inter_shift,1)) . $x . (substr($MUT_string,$a,1));
			}
		}
	}
	return $m_count;
}


##############################################################################
########################### PRINT FUNCTIONS ##################################
##############################################################################

#PRINT 1# - Function for print sequence into output file
#Return: 0
sub sequence_print
{
	#1 Variables
	my $name_file = shift;		# Name of output file
	my $name_seq = shift;		# Head of sequence
	my $sequence = shift;		# Sequence
	my $freq = shift;			# Number nucleotides/aminoacids per line
	my $i = 0;
	my $part = "";
	
	#2 Open FILE
	open( my $FILE, '>>', $name_file);
	
	#3 Print data into file
	print $FILE ">", $name_seq, "\n";
	for($i=0; $i<(length($sequence)); $i+=$freq)
	{
		$part = substr($sequence,$i,$freq);
		print $FILE $part, "\n";
	}

	#4 Close file and return
	close $FILE;
	return 0;
}

#PRINT 2# - Function for print record into file
#Return: 0
sub mutation_info_print
{
	#1 Variables
	my $n_gene = shift;
	my $n_transcript = shift;
	my $position = shift;
	my $WT = shift;
	my $MUT = shift;

	#2 Open FILE
	open(LOG, '>>', $file_name_log);

	#3 Print data into file
	print LOG $n_gene, "\t", $n_transcript, "\t", $position, "\t", $WT, "\t", $MUT, "\n";

	#4 Close file and return
	close LOG;
	return 0;
}

#PRINT 3# - Function for print help message - function()
#Return: 0
sub print_usage 
{
	print STDERR <<EOF
==================================================================================================
mut_pep_extract.pl - Software for translation sequences into protein based on mutation info
==================================================================================================

Usage:
	mut_pep_extract.pl [required params] [other params]

Required params:
	--input_reference : file : Input file with genome reference in fasta format
	--input_GTF       : file : Input file with GTF annotation
	--input_MUT       : file : Input file with mutation info with defined format

Other params:
	--help            :        : Prints the help message and exit.
	--out_WT_nucl     : string : Output file with nucleotide sequences from WT
	--out_MUT_nucl    : string : Output file with nucleotide sequences after mutation
	--out_WT_pep      : string : Output file with WT peptide sequence
	--out_MUT_pep     : string : Output file with MUT peptide sequence
	--out_log         : string : Output file with logs
 
EOF
}

#PRINT 4# - Function for reseting files
sub files_reset
{
	#1 Log file reset
	open(RES, '>', $file_name_log);
	close RES;

	#2 Error logfile reset
	open(ERRLOG, '>', "error.log");

	#2 Output files reset
	open(RES, '>', $file_name_WT_nucl);
	close RES;
	open(RES, '>', $file_name_WT_pep);
	close RES;
	open(RES, '>', $file_name_MUT_nucl);
	close RES;
	open(RES, '>', $file_name_MUT_pep);
	close RES;

	#3 Return function
	return 0;
}
